import Acl from 'acljs'
import { actionExists, getPermissions, Permissions } from './PermissionManager'
import { createImplement, invokeInterface } from 'sunny-js/util/module'
import { splitPath } from 'sunny-js/util/path'
import { RouteProps } from 'react-router'

export interface OperationProps {
  // 当前路由所属资源（resource）
  resource?: string
  // 当前路由组件对应的特权（privilege）
  privilege?: string
}

// 定义模块接口
interface ModuleInterfaces {
  getRole?: () => string
  defaultPrivilege?: string
  getOperation?: (route: RouteProps) => Required<OperationProps>
}
const iACL = {
  defaultPrivilege: 'view'
} as ModuleInterfaces
export const implement = createImplement(iACL)

export interface RouteLike {
  path: string,
  [key: string]: any
}
export type ACLRouteProps  = OperationProps & RouteLike

interface ACL {
  isAllowed: (role: string, resource: string, privilege: string) => boolean | never
}
let _acl: null | ACL = null
/**
 * 单例模式创建ACL实例，并初始化好权限
 * 该方法可以重复调用
 */
function getACLInstance (permissions: Permissions): ACL {
  return _acl = _acl || new Acl(permissions)
}

export const getRole = invokeInterface(
  () => iACL.getRole,
  () => 'guest'
)
export const getOperation = invokeInterface(
  () => iACL.getOperation,
  (route: ACLRouteProps) => {
    // 先将path分割成片段
    const path = route.path
    const segments = splitPath(path)

    // 非后台管理类的path默认privilege为view（查看）
    let privilege: string = actionExists(path) ? segments.pop() : iACL.defaultPrivilege
    // 优先使用路由配置中的相关字段
    privilege = route.privilege || privilege

    let resource: string = segments.pop()
    resource = route.resource || resource

    return { resource, privilege }
  }
)

// 权限验证
export function isAllowed (route: ACLRouteProps) {
  return getPermissions()
    .then(permissions => getACLInstance(permissions))
    .then(acl => {
      let allowed = false
      const role = getRole()
      const { resource, privilege } = getOperation(route)
      try {
        allowed = acl.isAllowed(role, resource, privilege)
      } catch (e) {
        console.error(e)
      }
      console.debug(
        role,
        allowed ? 'is' : 'is not',
        'allowed to',
        (privilege||'?') + '(privilege)', (resource||'?') + '(resource)')
      return allowed ? Promise.resolve() : Promise.reject()
    })
    .catch(e => {
      console.error(e)
      return Promise.reject()
    })
}
