import { isStr } from 'sunny-js/util/string'
import { splitPath } from 'sunny-js/util/path'
import { memoize } from 'sunny-js/util/function'
import { ACLRouteProps } from './AccessControl'
import { createImplement } from 'sunny-js/util/module'
import { invokeInterface } from 'sunny-js/util/module'
import { loadFromStorage, saveToSession } from 'sunny-js/util/storage'
import { isPlainObj } from 'sunny-js/util/object'

// 定义模块接口
interface ModuleInterfaces {
  // 自定义读取权限配置信息接口。不定义将使用默认权限配置
  AjaxPermissions?: () => Promise<Permissions>
  // 如果你想从所有路由配置中自动收集资源列表，可以定义该接口
  getRoutes?: () => ACLRouteProps[]
}
const iPM = {} as ModuleInterfaces
export const implement = createImplement(iPM)

interface Role {
  name: string
  parent: string
}

export type Resource = Role

export interface Rule {
  access: string
  role: string
  privileges: string[] | null
  resources: string[] | null
}

export interface Permissions {
  roles: Role[]
  resources: Resource[]
  rules: Rule[]
}

/**
 * 预定义对资源进行操作的相关关键词
 * 可以由此来判定path是否已经包含了内置privilege信息
 *
 * 用途：
 *  - 辅助识别或区分后台管理类应用路由的path
 */
const builtInActions = [
  // 查看类关键词
  'view', 'index', 'list', 'detail', 'read',
  // 操作类关键词
  'add', 'create', 'edit', 'del',
  // 导入导出
  'down', 'save', 'load', 'export', 'import'
]

/**
 * 补充缺失的Privileges为{@link actionExists}函数判断提供参考
 */
export function addMissingPrivileges (privileges: Rule[] | string[]) {
  const pushIfNotExists = (rule: string) => {
    builtInActions.includes(rule) || builtInActions.push(rule)
  }
  // 遍历所有访问规则
  privileges.forEach(rule => {
    if (isStr(rule)) {
      pushIfNotExists(rule)
    } else if (Array.isArray(rule.privileges)) {
      // 遍历当前访问规则，尝试加入新的privilege
      (rule as Rule).privileges.forEach(privilege => pushIfNotExists(privilege))
    }
  })
}

/**
 * 尝试判定预定义的Privileges是否至少有一个在route path中
 * 这样就可以辅助{@link getOperation}函数能够在route path中正常获取resource、privilege信息
 */
export const actionExists: (path: string) => boolean = memoize((path) => {
  return builtInActions.some(keyword => path.indexOf(keyword) !== -1)
})
// 读取路由配置信息列表
export const getRoutes = invokeInterface(
  () => iPM.getRoutes,
  () => [] as ACLRouteProps[]
)
/**
 * 更新资源列表对象，以确保resources中的资源都是唯一的
 */
function updateResources (resources: { [name: string]: string }, mixedResources: string | Resource[]) {
  if (isStr(mixedResources)) {
    const path = mixedResources as string
    const segments = splitPath(path)
    if (actionExists(path)) {
      segments.pop()
    }
    segments.reduce((previousValue, currentValue) => {
      resources[currentValue] = resources[currentValue] || previousValue
      return currentValue
    }, undefined)
  } else if (Array.isArray(mixedResources)) {
    mixedResources.forEach(resource => {
      const { name, parent } = resource || {}
      if (name) {
        resources[name] = resources[name] || parent
      }
    })
  }
}
/**
 * 从类似路由paths中，生成ACL中的{@link Resource}
 * Note:
 * - resource必须保证唯一性
 */
function generateResourcesFromRoutes (existingResources?: Resource[]) {
  const objResources = {}
  getRoutes().forEach(route => {
    const path = route.path
    if (Array.isArray(path)) {
      path.forEach(val => updateResources(objResources, val))
    } else {
      updateResources(objResources, path)
    }
  })
  updateResources(objResources, existingResources)
  return Object.entries<string>(objResources).map(([name, parent]) => ({
    name,
    parent
  }))
}

/**
 * 补充缺失的Resources。尝试从路由配置信息生成resources
 */
export function addMissingResources (ACLPermissions: Permissions) {
  ACLPermissions.resources = generateResourcesFromRoutes(
    ACLPermissions.resources
  )
  return ACLPermissions
}

let _cachePermissions: Permissions = null
function cachePermissions (permissions?: Permissions): Permissions | void {
  return isPlainObj(permissions)
    ? saveToSession({ ACLPermissions: permissions })
    : (_cachePermissions = _cachePermissions || loadFromStorage('ACLPermissions'))
}
const AjaxPermissions = invokeInterface(
  () => iPM.AjaxPermissions,
  () => import('./permissions.default')
    .then(({ default: permissions }) => permissions)
)
export function getPermissions () {
  const cachedPermissions = cachePermissions()
  return cachedPermissions
    // 优先使用缓存
    ? Promise.resolve(cachedPermissions)
    : AjaxPermissions()
      .then(permissions => {
        // 确保在isAllowed方法调用之前，补充缺失的Privileges、Resources
        addMissingResources(permissions)
        addMissingPrivileges(permissions.rules)
        // 缓存下来
        cachePermissions(permissions)
        return permissions
      })
}
